#!@PERL@

# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of make-world
#
# make-world is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# make-world is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#       Free Software Foundation, Inc.
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package MakeWorld::Archive;

use strict;
use warnings;

use version 0.77;

use File::Basename;
use File::Temp qw( tempdir );
use File::Copy;
use File::Path;
use Cwd qw( realpath getcwd );
use File::Spec::Functions qw/ catdir /;

use Archive::Extract;

use MakeWorld::Global;
use MakeWorld::Log;

use Module::Pluggable require => 1, search_path => qw[ MakeWorld::Package ], except => qr/::(Unknown|Custom)$/;
my @PackageTypes =  __PACKAGE__->plugins();

sub new
{
    my $class = shift;

    my $opts = shift;

    my $self = bless { %$opts }, $class;

    return $self->_find( @_ );
}

sub base { $_[0]->{base} }
sub ae   { $_[0]->{ae} }
sub class { $_[0]->{class} }

sub info {

    my $self = shift;

    return {
	    file => $self->{file},
	    base => $self->base,
	    package => $self->{package},
	    archive_name => $self->{archive_name},
	    vstring => $self->{vstring}
	   };
}

sub _find
{
    my $self = shift;

    my ( $dirs ) = @_;

    my $name = $self->{name};
    my $archive_name = $self->{archive_name};

    my @files;
    my @candidates;
    my $maxcomps = 0;

    my $cwd = getcwd;

    msg( MakeWorld::Log::Debug, "searching for archive" );

    # get a list of candidate files, either from a supplied RE
    # or by using our default glob
    if ( defined $self->{matchre} )
    {
	for my $dir ( @$dirs )
	{
	    chdir( $dir )
	      or die( "unable to chdir to $dir\n" );
	    push @candidates, 
	      map { "$dir/$_" }
	      grep { /$self->{matchre}/ } glob("*");
	    chdir( $cwd )
	      or die( "unable to chdir to $cwd\n" );
	}
    }
    else
    {
	@candidates = map { glob("$_/$archive_name-[0-9]*") } @$dirs;
    }

  FILE:
    for my $file ( @candidates )
    {

	msg( MakeWorld::Log::Debug, "Testing possible archive: " . $file );

	my %info = ( file => $file,
		     package => $name,
		     archive_name => $archive_name );

	# First, let the package classes scan the file name and see if
	# they can identify and use it.  They should only return a
	# positive match if the filename is unique to them (e.g., not
	# just that they can handle .tar.gz files).  At this stage
	# they should not extract archive files
	my @fileparse;

	for my $pkgclass ( @PackageTypes ) {
	    if ( @fileparse = $pkgclass->matches_archive( $file ) ) {
		$info{class} = $pkgclass;
		@info{qw[ base path suffix]} = @fileparse;
		last;
	    }
	}

	# if that didn't work, see if Archive::Extract can handle it
	unless( $info{class} ) {
	    local $Archive::Extract::WARN = 0;
	    $info{ae} = Archive::Extract->new( archive => $file );

	    # unfortunately Archive::Extract doesn't split up the
	    # file name for us, and we need to get it without an extension
	    # it's those .tar.xxx extensions that are a pain
	    @info{qw[ base path suffix]} = fileparse( $file,
						      qr/((?:\.tar)?\.\w+)/
						      )
	}

	# only care about things we understand
	next FILE unless $info{class} || $info{ae};

	# what's in this thing?
	# if it's a symbolic link, file looks at that, so get the realpath
	$info{rfile} = realpath($file);

	# use the first method which
	# is applicable
	#
	#  1. regular expression specified via the versionre attribute
	#
	#  2. a package class get_version method if a package class has
	#  claimed this package
	#
	#  3. a default regular expression
	# this see assume everything after last '-' is
	# it
	if ( $self->{versionre} ) {
	    ( $info{version} ) = $info{base} =~ /$self->{versionre}/;
	}

	elsif ( $info{class} && $info{class}->can('get_version') ) {
	    $info{version} = $info{class}->get_version( $name,
							$info{base} );
	}

	else {
	    ( $info{version} ) = $info{base} =~ /^$archive_name-([^-].*)/;
	}


	# extract version components

	# provided regexp
	if ( $self->{version_extract} ) {
	    my $re = $self->{version_extract};
	    my @comp = $info{version} =~ $re;

	    $info{version} = join('.', grep { defined $_ } @comp[0,1,2] );
	    $info{version_stage}  = $comp[3] if defined $comp[3];
	    $info{version_stagev} = $comp[4] if defined $comp[4];
	}

	# class?
	elsif ( $info{class} && $info{class}->can('extract_version' ) ) {

	    @info{qw[ version version_stage version_stagev]}
	      = $info{class}->extract_version( $info{version} )
	}

	# default
	else {

	    my @comp;
	    if ( @comp = $info{version} =~ /(.*)[-\.]?(rc|b)(\d+)/ ) {

		@info{qw[ version version_stage version_stagev]}
		  = @comp
	    }
	}

	$info{version_stage}  //= 'zzzzz';
	$info{version_stagev} //= '99999';

	$info{vobj} = eval { version->declare( $info{version} ) }
	  or die( "$@: $info{version}\n" );
	$info{vstring} = $info{vobj}->normal . $info{version_stage} . $info{version_stagev};

	push @files, \%info;
    }

    die( "unable to find archive for $name\n" )
      unless @files;

    my %is_dup_file;

    # recreate versions, filling in empty components if necessary
    for my $file ( @files )
    {
	$is_dup_file{$file->{vstring}}++;
    }

    # now, sort by version and pluck the greatest one
    my $latest = ( sort { $a->{vobj} <=> $b->{vobj} || $a->{vstring} cmp $b->{vstring} } @files )[-1];
    die( "unable to find archive for $name\n" )
      unless defined $latest;


    if ( $is_dup_file{$latest->{vstring}} > 1 )
    {
	my @dups = grep { $_->{vstring} eq $latest->{vstring} } @files;
	die( "duplicate versions of $name: \n\t", 
	     join(qq{\n\t}, map { $_->{file} } @dups ), "\n" );
    }

    %$self = %$latest;

    msg( MakeWorld::Log::Debug, "found archive: " . $self->{file} );
    return $self;
}


#----------------------------------------------------------------#

sub extract
{
    my ( $self, $logfile, $extract )  = @_;

    my $base = $self->{base};
    my $file = $self->{file};

    my $cwd = getcwd;


    # directory where the archive is extracted
    my $extdir;

    # this is the name of the directory that'll get renamed as $base.
    # either $extdir or $extdir/subdir
    my $srcdir;

    if ( $extract )
    {
	eval {

	    # extract archive into its own directory.  this allows
	    # processing of archives which don't unpack into
	    # a properly named and versioned directory

	    unless ( $opt{dryrun} )
	    {
		$extdir = tempdir( DIR => $cwd );

		msg( MakeWorld::Log::Notice, "extracting $file" );
		msg( MakeWorld::Log::Info, "extracting to $extdir" );

		chdir( $extdir )
		  or die( "unable to chdir to $extdir\n" );

		eval {
		    my $rd = MakeWorld::IORedir->new( file => $logfile );
		    local $Archive::Extract::PREFER_BIN = 1;
		    $self->{ae}->extract()
		      or die( "error extracting $file\n" )
		  };

		if ( my $err = $@ ) {
		    die defined $logfile ? "$err: see $logfile" : $err;
		}

		# now, what's in the current directory?
		my @dircontents = glob('*');

		# only one thing and its a directory. cool.
		if ( 1 == @dircontents && -d $dircontents[0] )
		{
		    $srcdir = catdir( $extdir, $dircontents[0] );
		}

		# more than one entry. rename $extdir to $base
		else
		{
		    $srcdir = $extdir;
		}
	    }

	    chdir( $cwd );
	    # remove build directory if it exists
	    if ( -d $base )
	    {
		msg( MakeWorld::Log::Notice, "removing existing build in $cwd/$base" );
		rmtree( $base )
		  unless $opt{dryrun};
	    }

	    unless ( $opt{dryrun} )
	    {
		msg( MakeWorld::Log::Info, "moving $srcdir to $base" );
		move( $srcdir, $base )
		  or die( "error renaming $srcdir to $base\n" );
	    }
	};

	my $error = $@;
	chdir( $cwd );

	# remove extdir if it hasn't been renamed to $base
	rmtree( $extdir ) if defined $extdir && -d $extdir;

	die $@ if $@;
    }

    return catdir( $cwd, $base );
}

1;
